@@ -6,6 +6,7 @@ use crate::compile::warn_missing_py_init;
 use crate::module_writer::write_python_part;
 use crate::module_writer::WheelWriter;
 use crate::module_writer::{write_bin, write_bindings_module, write_cffi_module};
+use crate::python_interpreter::InterpreterKind;
 use crate::source_distribution::source_distribution;
 use crate::Metadata21;
 use crate::PyProjectToml;
@@ -14,6 +15,7 @@ use crate::Target;
 use anyhow::{anyhow, bail, Context, Result};
 use cargo_metadata::Metadata;
 use fs_err as fs;
+use std::borrow::Cow;
 use std::path::{Path, PathBuf};
 
 /// The way the rust code is used in the wheel
@@ -36,7 +38,7 @@ impl BridgeModel {
     /// Returns the name of the bindings crate
     pub fn unwrap_bindings(&self) -> &str {
         match self {
-            BridgeModel::Bindings(value) => &value,
+            BridgeModel::Bindings(value) => value,
             _ => panic!("Expected Bindings"),
         }
     }
@@ -73,22 +75,29 @@ pub enum ProjectLayout {
 
 impl ProjectLayout {
     /// Checks whether a python module exists besides Cargo.toml with the right name
-    pub fn determine(project_root: impl AsRef<Path>, module_name: &str) -> Result<ProjectLayout> {
+    pub fn determine(
+        project_root: impl AsRef<Path>,
+        module_name: &str,
+        py_src: Option<impl AsRef<Path>>,
+    ) -> Result<ProjectLayout> {
         // A dot in the module name means the extension module goes into the module folder specified by the path
         let parts: Vec<&str> = module_name.split('.').collect();
         let project_root = project_root.as_ref();
+        let python_root = py_src.map_or(Cow::Borrowed(project_root), |py_src| {
+            Cow::Owned(project_root.join(py_src))
+        });
         let (python_module, rust_module, extension_name) = if parts.len() > 1 {
             let mut rust_module = project_root.to_path_buf();
             rust_module.extend(&parts[0..parts.len() - 1]);
             (
-                project_root.join(parts[0]),
+                python_root.join(parts[0]),
                 rust_module,
                 parts[parts.len() - 1].to_string(),
             )
         } else {
             (
-                project_root.join(module_name),
-                project_root.join(module_name),
+                python_root.join(module_name),
+                python_root.join(module_name),
                 module_name.to_string(),
             )
         };
@@ -164,6 +173,8 @@ pub struct BuildContext {
     pub cargo_metadata: Metadata,
     /// Whether to use universal2 or use the native macOS tag (off)
     pub universal2: bool,
+    /// Build editable wheels
+    pub editable: bool,
 }
 
 /// The wheel file location and its Python version tag (e.g. `py3`).
@@ -182,9 +193,32 @@ impl BuildContext {
         let wheels = match &self.bridge {
             BridgeModel::Cffi => self.build_cffi_wheel()?,
             BridgeModel::Bin => self.build_bin_wheel()?,
-            BridgeModel::Bindings(_) => self.build_binding_wheels()?,
+            BridgeModel::Bindings(_) => self.build_binding_wheels(&self.interpreter)?,
             BridgeModel::BindingsAbi3(major, minor) => {
-                self.build_binding_wheel_abi3(*major, *minor)?
+                let cpythons: Vec<_> = self
+                    .interpreter
+                    .iter()
+                    .filter(|interp| interp.interpreter_kind == InterpreterKind::CPython)
+                    .cloned()
+                    .collect();
+                let pypys: Vec<_> = self
+                    .interpreter
+                    .iter()
+                    .filter(|interp| interp.interpreter_kind == InterpreterKind::PyPy)
+                    .cloned()
+                    .collect();
+                let mut built_wheels = Vec::new();
+                if !cpythons.is_empty() {
+                    built_wheels.extend(self.build_binding_wheel_abi3(&cpythons, *major, *minor)?);
+                }
+                if !pypys.is_empty() {
+                    println!(
+                        "⚠️ Warning: PyPy does not yet support abi3 so the build artifacts will be version-specific. \
+                        See https://foss.heptapod.net/pypy/pypy/-/issues/3397 for more information."
+                    );
+                    built_wheels.extend(self.build_binding_wheels(&pypys)?);
+                }
+                built_wheels
             }
         };
 
@@ -226,7 +260,7 @@ impl BuildContext {
             .map(|x| &x.target)
             .unwrap_or(&self.target);
 
-        let policy = auditwheel_rs(&artifact, target, platform_tag).context(
+        let policy = auditwheel_rs(artifact, target, platform_tag).context(
             if let Some(platform_tag) = platform_tag {
                 format!("Error ensuring {} compliance", platform_tag)
             } else {
@@ -236,6 +270,13 @@ impl BuildContext {
         Ok(policy)
     }
 
+    fn add_pth(&self, writer: &mut WheelWriter) -> Result<()> {
+        if self.editable {
+            writer.add_pth(&self.project_layout, &self.metadata21)?;
+        }
+        Ok(())
+    }
+
     fn write_binding_wheel_abi3(
         &self,
         artifact: &Path,
@@ -243,7 +284,9 @@ impl BuildContext {
         major: u8,
         min_minor: u8,
     ) -> Result<BuiltWheelMetadata> {
-        let platform = self.target.get_platform_tag(platform_tag, self.universal2);
+        let platform = self
+            .target
+            .get_platform_tag(platform_tag, self.universal2)?;
         let tag = format!("cp{}{}-abi3-{}", major, min_minor, platform);
 
         let mut writer = WheelWriter::new(&tag, &self.out, &self.metadata21, &[tag.clone()])?;
@@ -255,10 +298,12 @@ impl BuildContext {
             artifact,
             None,
             &self.target,
-            false,
+            self.editable,
         )
         .context("Failed to add the files to the wheel")?;
 
+        self.add_pth(&mut writer)?;
+
         let wheel_path = writer.finish()?;
         Ok((wheel_path, format!("cp{}{}", major, min_minor)))
     }
@@ -267,13 +312,14 @@ impl BuildContext {
     /// for it
     pub fn build_binding_wheel_abi3(
         &self,
+        interpreters: &[PythonInterpreter],
         major: u8,
         min_minor: u8,
     ) -> Result<Vec<BuiltWheelMetadata>> {
         let mut wheels = Vec::new();
         // On windows, we have picked an interpreter to set the location of python.lib,
         // otherwise it's none
-        let python_interpreter = self.interpreter.get(0);
+        let python_interpreter = interpreters.get(0);
         let artifact = self.compile_cdylib(
             python_interpreter,
             Some(self.project_layout.extension_name()),
@@ -299,7 +345,7 @@ impl BuildContext {
         artifact: &Path,
         platform_tag: PlatformTag,
     ) -> Result<BuiltWheelMetadata> {
-        let tag = python_interpreter.get_tag(platform_tag, self.universal2);
+        let tag = python_interpreter.get_tag(platform_tag, self.universal2)?;
 
         let mut writer = WheelWriter::new(&tag, &self.out, &self.metadata21, &[tag.clone()])?;
 
@@ -307,13 +353,15 @@ impl BuildContext {
             &mut writer,
             &self.project_layout,
             &self.module_name,
-            &artifact,
-            Some(&python_interpreter),
+            artifact,
+            Some(python_interpreter),
             &self.target,
-            false,
+            self.editable,
         )
         .context("Failed to add the files to the wheel")?;
 
+        self.add_pth(&mut writer)?;
+
         let wheel_path = writer.finish()?;
         Ok((
             wheel_path,
@@ -328,15 +376,17 @@ impl BuildContext {
     /// and silently ignores all non-existent python versions.
     ///
     /// Runs [auditwheel_rs()] if not deactivated
-    pub fn build_binding_wheels(&self) -> Result<Vec<BuiltWheelMetadata>> {
+    pub fn build_binding_wheels(
+        &self,
+        interpreters: &[PythonInterpreter],
+    ) -> Result<Vec<BuiltWheelMetadata>> {
         let mut wheels = Vec::new();
-        for python_interpreter in &self.interpreter {
+        for python_interpreter in interpreters {
             let artifact = self.compile_cdylib(
-                Some(&python_interpreter),
+                Some(python_interpreter),
                 Some(self.project_layout.extension_name()),
             )?;
-            let policy =
-                self.auditwheel(Some(&python_interpreter), &artifact, self.platform_tag)?;
+            let policy = self.auditwheel(Some(python_interpreter), &artifact, self.platform_tag)?;
             let (wheel_path, tag) =
                 self.write_binding_wheel(python_interpreter, &artifact, policy.platform_tag())?;
             println!(
@@ -364,7 +414,7 @@ impl BuildContext {
         python_interpreter: Option<&PythonInterpreter>,
         extension_name: Option<&str>,
     ) -> Result<PathBuf> {
-        let artifacts = compile(&self, python_interpreter, &self.bridge)
+        let artifacts = compile(self, python_interpreter, &self.bridge)
             .context("Failed to build a native library through cargo")?;
 
         let artifact = artifacts.get("cdylib").cloned().ok_or_else(|| {
@@ -389,21 +439,23 @@ impl BuildContext {
     ) -> Result<BuiltWheelMetadata> {
         let (tag, tags) = self
             .target
-            .get_universal_tags(platform_tag, self.universal2);
+            .get_universal_tags(platform_tag, self.universal2)?;
 
-        let mut builder = WheelWriter::new(&tag, &self.out, &self.metadata21, &tags)?;
+        let mut writer = WheelWriter::new(&tag, &self.out, &self.metadata21, &tags)?;
 
         write_cffi_module(
-            &mut builder,
+            &mut writer,
             &self.project_layout,
             self.manifest_path.parent().unwrap(),
             &self.module_name,
-            &artifact,
+            artifact,
             &self.interpreter[0].executable,
-            false,
+            self.editable,
         )?;
 
-        let wheel_path = builder.finish()?;
+        self.add_pth(&mut writer)?;
+
+        let wheel_path = writer.finish()?;
         Ok((wheel_path, "py3".to_string()))
     }
 
@@ -414,6 +466,19 @@ impl BuildContext {
         let policy = self.auditwheel(None, &artifact, self.platform_tag)?;
         let (wheel_path, tag) = self.write_cffi_wheel(&artifact, policy.platform_tag())?;
 
+        // Warn if cffi isn't specified in the requirements
+        if !self
+            .metadata21
+            .requires_dist
+            .iter()
+            .any(|dep| dep.to_ascii_lowercase().starts_with("cffi"))
+        {
+            eprintln!(
+                "⚠️  Warning: missing cffi package dependency, please add it to pyproject.toml. \
+                e.g: `dependencies = [\"cffi\"]`. This will become an error."
+            );
+        }
+
         println!("📦 Built wheel to {}", wheel_path.display());
         wheels.push((wheel_path, tag));
 
@@ -427,13 +492,13 @@ impl BuildContext {
     ) -> Result<BuiltWheelMetadata> {
         let (tag, tags) = self
             .target
-            .get_universal_tags(platform_tag, self.universal2);
+            .get_universal_tags(platform_tag, self.universal2)?;
 
         if !self.metadata21.scripts.is_empty() {
             bail!("Defining entrypoints and working with a binary doesn't mix well");
         }
 
-        let mut builder = WheelWriter::new(&tag, &self.out, &self.metadata21, &tags)?;
+        let mut writer = WheelWriter::new(&tag, &self.out, &self.metadata21, &tags)?;
 
         match self.project_layout {
             ProjectLayout::Mixed {
@@ -441,8 +506,10 @@ impl BuildContext {
                 ref extension_name,
                 ..
             } => {
-                write_python_part(&mut builder, python_module, extension_name)
-                    .context("Failed to add the python module to the package")?;
+                if !self.editable {
+                    write_python_part(&mut writer, python_module, extension_name)
+                        .context("Failed to add the python module to the package")?;
+                }
             }
             ProjectLayout::PureRust { .. } => {}
         }
@@ -452,9 +519,11 @@ impl BuildContext {
         let bin_name = artifact
             .file_name()
             .expect("Couldn't get the filename from the binary produced by cargo");
-        write_bin(&mut builder, &artifact, &self.metadata21, bin_name)?;
+        write_bin(&mut writer, artifact, &self.metadata21, bin_name)?;
 
-        let wheel_path = builder.finish()?;
+        self.add_pth(&mut writer)?;
+
+        let wheel_path = writer.finish()?;
         Ok((wheel_path, "py3".to_string()))
     }
 
@@ -463,7 +532,7 @@ impl BuildContext {
     /// Runs [auditwheel_rs()] if not deactivated
     pub fn build_bin_wheel(&self) -> Result<Vec<BuiltWheelMetadata>> {
         let mut wheels = Vec::new();
-        let artifacts = compile(&self, None, &self.bridge)
+        let artifacts = compile(self, None, &self.bridge)
             .context("Failed to build a native library through cargo")?;
 
         let artifact = artifacts
