@@ -4,171 +4,20 @@
 //! Run with --help for usage information
 
 use anyhow::{bail, Context, Result};
-#[cfg(feature = "upload")]
-use bytesize::ByteSize;
 use cargo_metadata::MetadataCommand;
-#[cfg(feature = "upload")]
-use configparser::ini::Ini;
 use fs_err as fs;
-#[cfg(feature = "human-panic")]
-use human_panic::setup_panic;
-#[cfg(feature = "password-storage")]
-use keyring::{Keyring, KeyringError};
+use maturin::GenerateProjectOptions;
 use maturin::{
-    develop, new_project, source_distribution, write_dist_info, BridgeModel, BuildOptions,
-    CargoToml, Metadata21, PathWriter, PlatformTag, PyProjectToml, PythonInterpreter, Target,
+    develop, init_project, new_project, source_distribution, write_dist_info, BridgeModel,
+    BuildOptions, CargoToml, Metadata21, PathWriter, PlatformTag, PyProjectToml, PythonInterpreter,
+    Target,
 };
+#[cfg(feature = "upload")]
+use maturin::{upload_ui, PublishOpt};
 use std::env;
 use std::io;
 use std::path::PathBuf;
 use structopt::StructOpt;
-#[cfg(feature = "upload")]
-use {
-    maturin::{upload, Registry, UploadError},
-    reqwest::Url,
-};
-
-/// Returns the password and a bool that states whether to ask for re-entering the password
-/// after a failed authentication
-///
-/// Precedence:
-/// 1. MATURIN_PASSWORD
-/// 2. keyring
-/// 3. stdin
-#[cfg(feature = "upload")]
-fn get_password(_username: &str) -> String {
-    if let Ok(password) = env::var("MATURIN_PASSWORD") {
-        return password;
-    };
-
-    #[cfg(feature = "keyring")]
-    {
-        let service = env!("CARGO_PKG_NAME");
-        let keyring = Keyring::new(&service, &_username);
-        if let Ok(password) = keyring.get_password() {
-            return password;
-        };
-    }
-
-    rpassword::prompt_password_stdout("Please enter your password: ").unwrap_or_else(|_| {
-        // So we need this fallback for pycharm on windows
-        let mut password = String::new();
-        io::stdin()
-            .read_line(&mut password)
-            .expect("Failed to read line");
-        password.trim().to_string()
-    })
-}
-
-#[cfg(feature = "upload")]
-fn get_username() -> String {
-    println!("Please enter your username:");
-    let mut line = String::new();
-    io::stdin().read_line(&mut line).unwrap();
-    line.trim().to_string()
-}
-
-#[cfg(feature = "upload")]
-fn load_pypirc() -> Ini {
-    let mut config = Ini::new();
-    if let Some(mut config_path) = dirs::home_dir() {
-        config_path.push(".pypirc");
-        if let Ok(pypirc) = fs::read_to_string(config_path.as_path()) {
-            let _ = config.read(pypirc);
-        }
-    }
-    config
-}
-
-#[cfg(feature = "upload")]
-fn load_pypi_cred_from_config(config: &Ini, registry_name: &str) -> Option<(String, String)> {
-    if let (Some(username), Some(password)) = (
-        config.get(registry_name, "username"),
-        config.get(registry_name, "password"),
-    ) {
-        return Some((username, password));
-    }
-    None
-}
-
-#[cfg(feature = "upload")]
-fn resolve_pypi_cred(
-    opt: &PublishOpt,
-    config: &Ini,
-    registry_name: Option<&str>,
-) -> (String, String) {
-    // API token from environment variable takes priority
-    if let Ok(token) = env::var("MATURIN_PYPI_TOKEN") {
-        return ("__token__".to_string(), token);
-    }
-
-    if let Some((username, password)) =
-        registry_name.and_then(|name| load_pypi_cred_from_config(config, name))
-    {
-        println!("🔐 Using credential in pypirc for upload");
-        return (username, password);
-    }
-
-    // fallback to username and password
-    let username = opt.username.clone().unwrap_or_else(get_username);
-    let password = match opt.password {
-        Some(ref password) => password.clone(),
-        None => get_password(&username),
-    };
-
-    (username, password)
-}
-
-#[cfg(feature = "upload")]
-/// Asks for username and password for a registry account where missing.
-fn complete_registry(opt: &PublishOpt) -> Result<Registry> {
-    // load creds from pypirc if found
-    let pypirc = load_pypirc();
-    let (register_name, registry_url) =
-        if !opt.registry.starts_with("http://") && !opt.registry.starts_with("https://") {
-            if let Some(url) = pypirc.get(&opt.registry, "repository") {
-                (Some(opt.registry.as_str()), url)
-            } else {
-                bail!(
-                    "Failed to get registry {} in .pypirc. \
-                    Note: Your index didn't start with http:// or https://, \
-                    which is required for non-pypirc indices.",
-                    opt.registry
-                );
-            }
-        } else if opt.registry == "https://upload.pypi.org/legacy/" {
-            (Some("pypi"), opt.registry.clone())
-        } else {
-            (None, opt.registry.clone())
-        };
-    let (username, password) = resolve_pypi_cred(opt, &pypirc, register_name);
-    let registry = Registry::new(username, password, Url::parse(&registry_url)?);
-
-    Ok(registry)
-}
-
-/// An account with a registry, possibly incomplete
-#[derive(Debug, StructOpt)]
-struct PublishOpt {
-    #[structopt(
-        short = "r",
-        long = "repository-url",
-        default_value = "https://upload.pypi.org/legacy/"
-    )]
-    /// The url of registry where the wheels are uploaded to
-    registry: String,
-    #[structopt(short, long)]
-    /// Username for pypi or your custom registry
-    username: Option<String>,
-    #[structopt(short, long)]
-    /// Password for pypi or your custom registry. Note that you can also pass the password
-    /// through MATURIN_PASSWORD
-    password: Option<String>,
-    /// Continue uploading files if one already exists.
-    /// (Only valid when uploading to PyPI. Other implementations may not support this.)
-    #[structopt(long = "skip-existing")]
-    skip_existing: bool,
-}
 
 #[derive(Debug, StructOpt)]
 #[structopt(name = env!("CARGO_PKG_NAME"))]
@@ -270,20 +119,21 @@ enum Opt {
         #[structopt(short, long, parse(from_os_str))]
         out: Option<PathBuf>,
     },
+    /// Create a new cargo project in an existing directory
+    #[structopt(name = "init")]
+    InitProject {
+        /// Project path
+        path: Option<String>,
+        #[structopt(flatten)]
+        options: GenerateProjectOptions,
+    },
     /// Create a new cargo project
     #[structopt(name = "new")]
     NewProject {
         /// Project path
         path: String,
-        /// Set the resulting package name, defaults to the directory name
-        #[structopt(long)]
-        name: Option<String>,
-        /// Use mixed Rust/Python project layout
-        #[structopt(long)]
-        mixed: bool,
-        /// Which kind of bindings to use
-        #[structopt(short, long, possible_values = &["pyo3", "rust-cpython", "cffi", "bin"])]
-        bindings: Option<String>,
+        #[structopt(flatten)]
+        options: GenerateProjectOptions,
     },
     /// Uploads python packages to pypi
     ///
@@ -437,82 +287,6 @@ fn pep517(subcommand: Pep517Command) -> Result<()> {
     Ok(())
 }
 
-/// Handles authentication/keyring integration and retrying of the publish subcommand
-#[cfg(feature = "upload")]
-fn upload_ui(items: &[PathBuf], publish: &PublishOpt) -> Result<()> {
-    let registry = complete_registry(publish)?;
-
-    println!("🚀 Uploading {} packages", items.len());
-
-    for i in items {
-        let upload_result = upload(&registry, i);
-
-        match upload_result {
-            Ok(()) => (),
-            Err(UploadError::AuthenticationError) => {
-                println!("⛔ Username and/or password are wrong");
-
-                #[cfg(feature = "keyring")]
-                {
-                    // Delete the wrong password from the keyring
-                    let old_username = registry.username.clone();
-                    let keyring = Keyring::new(&env!("CARGO_PKG_NAME"), &old_username);
-                    match keyring.delete_password() {
-                        Ok(()) => {
-                            println!("🔑 Removed wrong password from keyring")
-                        }
-                        Err(KeyringError::NoPasswordFound) | Err(KeyringError::NoBackendFound) => {}
-                        Err(err) => {
-                            eprintln!("⚠️ Warning: Failed to remove password from keyring: {}", err)
-                        }
-                    }
-                }
-
-                bail!("Username and/or password are wrong");
-            }
-            Err(err) => {
-                let filename = i.file_name().unwrap_or_else(|| i.as_os_str());
-                if let UploadError::FileExistsError(_) = err {
-                    if publish.skip_existing {
-                        println!(
-                            "⚠️ Note: Skipping {:?} because it appears to already exist",
-                            filename
-                        );
-                        continue;
-                    }
-                }
-                let filesize = fs::metadata(&i)
-                    .map(|x| ByteSize(x.len()).to_string())
-                    .unwrap_or_else(|e| format!("Failed to get the filesize of {:?}: {}", &i, e));
-                return Err(err)
-                    .context(format!("💥 Failed to upload {:?} ({})", filename, filesize));
-            }
-        }
-    }
-
-    println!("✨ Packages uploaded successfully");
-
-    #[cfg(feature = "keyring")]
-    {
-        // We know the password is correct, so we can save it in the keyring
-        let username = registry.username.clone();
-        let keyring = Keyring::new(&env!("CARGO_PKG_NAME"), &username);
-        let password = registry.password.clone();
-        match keyring.set_password(&password) {
-            Ok(()) => {}
-            Err(KeyringError::NoBackendFound) => {}
-            Err(err) => {
-                eprintln!(
-                    "⚠️ Warning: Failed to store the password in the keyring: {:?}",
-                    err
-                );
-            }
-        }
-    }
-
-    Ok(())
-}
-
 fn run() -> Result<()> {
     #[cfg(feature = "log")]
     pretty_env_logger::init();
@@ -636,12 +410,8 @@ fn run() -> Result<()> {
             .context("Failed to build source distribution")?;
         }
         Opt::Pep517(subcommand) => pep517(subcommand)?,
-        Opt::NewProject {
-            path,
-            name,
-            mixed,
-            bindings,
-        } => new_project(path, name, mixed, bindings)?,
+        Opt::InitProject { path, options } => init_project(path, options)?,
+        Opt::NewProject { path, options } => new_project(path, options)?,
         #[cfg(feature = "upload")]
         Opt::Upload { publish, files } => {
             if files.is_empty() {
@@ -662,7 +432,7 @@ fn run() -> Result<()> {
 fn main() {
     #[cfg(feature = "human-panic")]
     {
-        setup_panic!();
+        human_panic::setup_panic!();
     }
 
     if let Err(e) = run() {
