@@ -356,6 +356,8 @@ impl Metadata21 {
             gui_scripts: HashMap::new(),
             entry_points: HashMap::new(),
         };
+
+        let manifest_path = manifest_path.as_ref();
         metadata.merge_pyproject_toml(manifest_path)?;
         Ok(metadata)
     }
@@ -391,7 +393,7 @@ impl Metadata21 {
             }
         };
 
-        add_option("Summary", &self.summary.as_deref().map(fold_header));
+        add_option("Summary", &self.summary);
         add_option("Keywords", &self.keywords);
         add_option("Home-Page", &self.home_page);
         add_option("Download-URL", &self.download_url);
@@ -486,7 +488,7 @@ fn fold_header(text: &str) -> String {
         if line.is_empty() {
             result.push('\t');
         } else {
-            result.push_str(&line);
+            result.push_str(line);
         }
     }
 
@@ -776,4 +778,20 @@ mod test {
         let pkginfo: Result<python_pkginfo::Metadata, _> = content.parse();
         assert!(pkginfo.is_ok());
     }
+
+    #[test]
+    fn test_merge_metadata_from_pyproject_toml_with_customized_python_source_dir() {
+        let cargo_toml_str =
+            fs_err::read_to_string("test-crates/pyo3-mixed-py-subdir/Cargo.toml").unwrap();
+        let cargo_toml: CargoToml = toml::from_str(&cargo_toml_str).unwrap();
+        let metadata =
+            Metadata21::from_cargo_toml(&cargo_toml, "test-crates/pyo3-mixed-py-subdir").unwrap();
+        // defined in Cargo.toml
+        assert_eq!(
+            metadata.summary,
+            Some("Implements a dummy function combining rust and python".to_string())
+        );
+        // defined in pyproject.toml
+        assert_eq!(metadata.scripts["get_42"], "pyo3_mixed_py_subdir:get_42");
+    }
 }
