@@ -1,5 +1,4 @@
 use crate::auditwheel::PlatformTag;
-use crate::cross_compile::is_cross_compiling;
 use crate::{BridgeModel, Target};
 use anyhow::{bail, format_err, Context, Result};
 use regex::Regex;
@@ -294,6 +293,11 @@ pub struct PythonInterpreter {
     ///
     /// Note that this can be `None` when cross compiling
     pub platform: Option<String>,
+    /// Is this interpreter runnable
+    ///
+    /// When cross compile the target interpreter isn't runnable,
+    /// and it's `executable` is empty
+    pub runnable: bool,
 }
 
 /// Returns the abiflags that are assembled through the message, with some
@@ -309,7 +313,7 @@ fn fun_with_abiflags(
 ) -> Result<String> {
     if bridge != &BridgeModel::Cffi
         && target.get_python_os() != message.system
-        && !is_cross_compiling(target)?
+        && !target.cross_compiling()
     {
         bail!(
             "platform.system() in python, {}, and the rust target, {:?}, don't match ಠ_ಠ",
@@ -356,17 +360,22 @@ impl PythonInterpreter {
     /// Don't ask me why or how, this is just what setuptools uses so I'm also going to use
     ///
     /// If abi3 is true, cpython wheels use the generic abi3 with the given version as minimum
-    pub fn get_tag(&self, platform_tag: PlatformTag, universal2: bool) -> String {
-        let platform = if self.target.is_windows() {
-            // Restrict `sysconfig.get_platform()` usage to Windows only for now
-            // so we don't need to deal with macOS deployment target
-            self.platform
-                .clone()
-                .unwrap_or_else(|| self.target.get_platform_tag(platform_tag, universal2))
+    pub fn get_tag(&self, platform_tag: PlatformTag, universal2: bool) -> Result<String> {
+        // Restrict `sysconfig.get_platform()` usage to Windows and non-portable Linux only for now
+        // so we don't need to deal with macOS deployment target
+        let use_sysconfig_platform = self.target.is_windows()
+            || (self.target.is_linux() && !platform_tag.is_portable())
+            || self.target.is_illumos();
+        let platform = if use_sysconfig_platform {
+            if let Some(platform) = self.platform.clone() {
+                platform
+            } else {
+                self.target.get_platform_tag(platform_tag, universal2)?
+            }
         } else {
-            self.target.get_platform_tag(platform_tag, universal2)
+            self.target.get_platform_tag(platform_tag, universal2)?
         };
-        match self.interpreter_kind {
+        let tag = match self.interpreter_kind {
             InterpreterKind::CPython => {
                 if self.target.is_unix() {
                     format!(
@@ -401,7 +410,8 @@ impl PythonInterpreter {
                     platform = platform,
                 )
             }
-        }
+        };
+        Ok(tag)
     }
 
     /// Adds the ext_suffix we read from python or know (.pyd/.abi3.so) and adds it to the base name
@@ -496,21 +506,17 @@ impl PythonInterpreter {
         ))?;
 
         let platform = if message.platform.starts_with("macosx") {
-            // macOS contains system version in platform name so need special handle
-            let mut parts = message.platform.splitn(3, '-');
-            let err_ctx = "Invalid platform string from `sysconfig.get_platform()`";
-            let prefix = parts.next().context(err_ctx)?;
-            let base_version = parts.next().context(err_ctx)?;
-            let suffix = parts.next().context(err_ctx)?;
-            let (major_ver, minor_ver) = base_version.split_once('.').context(err_ctx)?;
-            let major_ver: u64 = major_ver.parse()?;
-            let minor_ver = if major_ver > 10 { "0" } else { minor_ver };
-            // FIXME: We should also take MACOSX_DEPLOYMENT_TARGET env var into account
-            format!("{}_{}_{}_{}", prefix, major_ver, minor_ver, suffix)
+            // We don't use platform from sysconfig on macOS
+            None
         } else {
-            message.platform
+            Some(
+                message
+                    .platform
+                    .to_lowercase()
+                    .replace('-', "_")
+                    .replace('.', "_"),
+            )
         };
-        let platform = platform.to_lowercase().replace('-', "_").replace('.', "_");
 
         Ok(Some(PythonInterpreter {
             major: message.major,
@@ -524,7 +530,8 @@ impl PythonInterpreter {
             interpreter_kind: interpreter,
             abi_tag: message.abi_tag,
             libs_dir: PathBuf::from(message.base_prefix).join("libs"),
-            platform: Some(platform),
+            platform,
+            runnable: true,
         }))
     }
 
@@ -590,6 +597,9 @@ impl PythonInterpreter {
 
     /// Run a python script using this Python interpreter.
     pub fn run_script(&self, script: &str) -> Result<String> {
+        if !self.runnable {
+            bail!("This {} isn't runnable", self);
+        }
         let out = Command::new(&self.executable)
             .env("PYTHONIOENCODING", "utf-8")
             .stdin(Stdio::piped())
@@ -630,14 +640,22 @@ impl PythonInterpreter {
 
 impl fmt::Display for PythonInterpreter {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        write!(
-            f,
-            "{} {}.{}{} at {}",
-            self.interpreter_kind,
-            self.major,
-            self.minor,
-            self.abiflags,
-            self.executable.display()
-        )
+        if self.runnable {
+            write!(
+                f,
+                "{} {}.{}{} at {}",
+                self.interpreter_kind,
+                self.major,
+                self.minor,
+                self.abiflags,
+                self.executable.display()
+            )
+        } else {
+            write!(
+                f,
+                "cross compiling target {} {}.{}{}",
+                self.interpreter_kind, self.major, self.minor, self.abiflags,
+            )
+        }
     }
 }
