@@ -1,3 +1,4 @@
+use crate::cross_compile::is_cross_compiling;
 use crate::python_interpreter::InterpreterKind;
 use crate::{PlatformTag, PythonInterpreter};
 use anyhow::{bail, format_err, Context, Result};
@@ -19,6 +20,8 @@ enum Os {
     FreeBsd,
     NetBsd,
     OpenBsd,
+    Dragonfly,
+    Illumos,
 }
 
 impl fmt::Display for Os {
@@ -30,6 +33,8 @@ impl fmt::Display for Os {
             Os::FreeBsd => write!(f, "FreeBSD"),
             Os::NetBsd => write!(f, "NetBSD"),
             Os::OpenBsd => write!(f, "OpenBSD"),
+            Os::Dragonfly => write!(f, "DragonFly"),
+            Os::Illumos => write!(f, "Illumos"),
         }
     }
 }
@@ -82,6 +87,8 @@ fn get_supported_architectures(os: &Os) -> Vec<Arch> {
             Arch::X86_64,
         ],
         Os::OpenBsd => vec![Arch::X86, Arch::X86_64, Arch::Aarch64],
+        Os::Dragonfly => vec![Arch::X86_64],
+        Os::Illumos => vec![Arch::X86_64],
     }
 }
 
@@ -92,6 +99,7 @@ pub struct Target {
     arch: Arch,
     env: Environment,
     triple: String,
+    cross_compiling: bool,
 }
 
 impl Target {
@@ -121,6 +129,8 @@ impl Target {
             target_lexicon::OperatingSystem::Netbsd => Os::NetBsd,
             target_lexicon::OperatingSystem::Freebsd => Os::FreeBsd,
             target_lexicon::OperatingSystem::Openbsd => Os::OpenBsd,
+            target_lexicon::OperatingSystem::Dragonfly => Os::Dragonfly,
+            target_lexicon::OperatingSystem::Illumos => Os::Illumos,
             unsupported => bail!("The operating system {:?} is not supported", unsupported),
         };
 
@@ -139,17 +149,20 @@ impl Target {
             bail!("{} is not supported on {}", arch, os);
         }
 
-        Ok(Target {
+        let mut target = Target {
             os,
             arch,
             env: platform.environment,
             triple,
-        })
+            cross_compiling: false,
+        };
+        target.cross_compiling = is_cross_compiling(&target)?;
+        Ok(target)
     }
 
     /// Returns the platform part of the tag for the wheel name
-    pub fn get_platform_tag(&self, platform_tag: PlatformTag, universal2: bool) -> String {
-        match (&self.os, &self.arch) {
+    pub fn get_platform_tag(&self, platform_tag: PlatformTag, universal2: bool) -> Result<String> {
+        let tag = match (&self.os, &self.arch) {
             // FreeBSD
             (Os::FreeBsd, Arch::X86_64)
             | (Os::FreeBsd, Arch::Aarch64)
@@ -163,10 +176,7 @@ impl Target {
             | (Os::OpenBsd, Arch::X86)
             | (Os::OpenBsd, Arch::X86_64)
             | (Os::OpenBsd, Arch::Aarch64) => {
-                let info = match PlatformInfo::new() {
-                    Ok(info) => info,
-                    Err(error) => panic!("{}", error),
-                };
+                let info = PlatformInfo::new()?;
                 let release = info.release().replace(".", "_").replace("-", "_");
                 let arch = match self.arch {
                     Arch::X86_64 => "amd64",
@@ -185,32 +195,87 @@ impl Target {
                     arch
                 )
             }
+            (Os::Dragonfly, Arch::X86_64) => {
+                let info = PlatformInfo::new()?;
+                let release = info.release().replace(".", "_").replace("-", "_");
+                format!(
+                    "{}_{}_{}",
+                    self.os.to_string().to_ascii_lowercase(),
+                    release.to_ascii_lowercase(),
+                    "x86_64"
+                )
+            }
+            (Os::Illumos, Arch::X86_64) => {
+                let info = PlatformInfo::new()?;
+                let mut release = info.release().replace(".", "_").replace("-", "_");
+                let mut arch = info.machine().replace(' ', "_").replace('/', "_");
+
+                let mut os = self.os.to_string().to_ascii_lowercase();
+                // See https://github.com/python/cpython/blob/46c8d915715aa2bd4d697482aa051fe974d440e1/Lib/sysconfig.py#L722-L730
+                if let Some((major, other)) = release.split_once('_') {
+                    let major_ver: u64 = major.parse().context("illumos major version is not a number")?;
+                    if major_ver >= 5 {
+                        // SunOS 5 == Solaris 2
+                        os = "solaris".to_string();
+                        release = format!("{}_{}", major_ver - 3, other);
+                        arch = format!("{}_64bit", arch);
+                    }
+                }
+                format!(
+                    "{}_{}_{}",
+                    os,
+                    release,
+                    arch
+                )
+            }
             (Os::Linux, _) => {
-                let mut tags = vec![format!("{}_{}", platform_tag, self.arch)];
+                let arch = if self.cross_compiling {
+                    self.arch.to_string()
+                } else {
+                    PlatformInfo::new()
+                        .map(|info| info.machine().into_owned())
+                        .unwrap_or_else(|_| self.arch.to_string())
+                };
+                let mut tags = vec![format!("{}_{}", platform_tag, arch)];
                 for alias in platform_tag.aliases() {
-                    tags.push(format!("{}_{}", alias, self.arch));
+                    tags.push(format!("{}_{}", alias, arch));
                 }
                 tags.join(".")
             }
             (Os::Macos, Arch::X86_64) => {
+                let ((x86_64_major, x86_64_minor), (arm64_major, arm64_minor)) = macosx_deployment_target(env::var("MACOSX_DEPLOYMENT_TARGET").ok().as_deref(), universal2)?;
                 if universal2 {
-                    "macosx_10_9_x86_64.macosx_11_0_arm64.macosx_10_9_universal2".to_string()
+                    format!(
+                        "macosx_{x86_64_major}_{x86_64_minor}_x86_64.macosx_{arm64_major}_{arm64_minor}_arm64.macosx_{x86_64_major}_{x86_64_minor}_universal2",
+                        x86_64_major = x86_64_major,
+                        x86_64_minor = x86_64_minor,
+                        arm64_major = arm64_major,
+                        arm64_minor = arm64_minor
+                    )
                 } else {
-                    "macosx_10_7_x86_64".to_string()
+                    format!("macosx_{}_{}_x86_64", x86_64_major, x86_64_minor)
                 }
             }
             (Os::Macos, Arch::Aarch64) => {
+                let ((x86_64_major, x86_64_minor), (arm64_major, arm64_minor)) = macosx_deployment_target(env::var("MACOSX_DEPLOYMENT_TARGET").ok().as_deref(), universal2)?;
                 if universal2 {
-                    "macosx_10_9_x86_64.macosx_11_0_arm64.macosx_10_9_universal2".to_string()
+                    format!(
+                        "macosx_{x86_64_major}_{x86_64_minor}_x86_64.macosx_{arm64_major}_{arm64_minor}_arm64.macosx_{x86_64_major}_{x86_64_minor}_universal2",
+                        x86_64_major = x86_64_major,
+                        x86_64_minor = x86_64_minor,
+                        arm64_major = arm64_major,
+                        arm64_minor = arm64_minor
+                    )
                 } else {
-                    "macosx_11_0_arm64".to_string()
+                    format!("macosx_{}_{}_arm64", arm64_major, arm64_minor)
                 }
             }
             (Os::Windows, Arch::X86) => "win32".to_string(),
             (Os::Windows, Arch::X86_64) => "win_amd64".to_string(),
             (Os::Windows, Arch::Aarch64) => "win_arm64".to_string(),
             (_, _) => panic!("unsupported target should not have reached get_platform_tag()"),
-        }
+        };
+        Ok(tag)
     }
 
     /// Returns the name python uses in `sys.platform` for this os
@@ -222,6 +287,8 @@ impl Target {
             Os::FreeBsd => "freebsd",
             Os::NetBsd => "netbsd",
             Os::OpenBsd => "openbsd",
+            Os::Dragonfly => "dragonfly",
+            Os::Illumos => "sunos",
         }
     }
 
@@ -257,7 +324,13 @@ impl Target {
     pub fn is_unix(&self) -> bool {
         match self.os {
             Os::Windows => false,
-            Os::Linux | Os::Macos | Os::FreeBsd | Os::NetBsd | Os::OpenBsd => true,
+            Os::Linux
+            | Os::Macos
+            | Os::FreeBsd
+            | Os::NetBsd
+            | Os::OpenBsd
+            | Os::Dragonfly
+            | Os::Illumos => true,
         }
     }
 
@@ -286,6 +359,11 @@ impl Target {
         self.os == Os::Windows
     }
 
+    /// Returns true if the current platform is illumos
+    pub fn is_illumos(&self) -> bool {
+        self.os == Os::Illumos
+    }
+
     /// Returns true if the current platform's target env is Musl
     pub fn is_musl_target(&self) -> bool {
         matches!(
@@ -297,12 +375,18 @@ impl Target {
         )
     }
 
+    /// Is cross compiling for this target
+    pub fn cross_compiling(&self) -> bool {
+        self.cross_compiling
+    }
+
     /// Returns the tags for the WHEEL file for cffi wheels
-    pub fn get_py3_tags(&self, platform_tag: PlatformTag, universal2: bool) -> Vec<String> {
-        vec![format!(
+    pub fn get_py3_tags(&self, platform_tag: PlatformTag, universal2: bool) -> Result<Vec<String>> {
+        let tags = vec![format!(
             "py3-none-{}",
-            self.get_platform_tag(platform_tag, universal2)
-        )]
+            self.get_platform_tag(platform_tag, universal2)?
+        )];
+        Ok(tags)
     }
 
     /// Returns the path to the python executable inside a venv
@@ -372,13 +456,13 @@ impl Target {
         &self,
         platform_tag: PlatformTag,
         universal2: bool,
-    ) -> (String, Vec<String>) {
+    ) -> Result<(String, Vec<String>)> {
         let tag = format!(
             "py3-none-{platform}",
-            platform = self.get_platform_tag(platform_tag, universal2)
+            platform = self.get_platform_tag(platform_tag, universal2)?
         );
-        let tags = self.get_py3_tags(platform_tag, universal2);
-        (tag, tags)
+        let tags = self.get_py3_tags(platform_tag, universal2)?;
+        Ok((tag, tags))
     }
 }
 
@@ -415,3 +499,65 @@ pub(crate) fn get_host_target() -> Result<String> {
         .to_string();
     Ok(host)
 }
+
+fn macosx_deployment_target(
+    deploy_target: Option<&str>,
+    universal2: bool,
+) -> Result<((usize, usize), (usize, usize))> {
+    let x86_64_default = if universal2 { (10, 9) } else { (10, 7) };
+    let arm64_default = (11, 0);
+    let mut x86_64_ver = x86_64_default;
+    let mut arm64_ver = arm64_default;
+    if let Some(deploy_target) = deploy_target {
+        let err_ctx = "MACOSX_DEPLOYMENT_TARGET is invalid";
+        let mut parts = deploy_target.split('.');
+        let major = parts.next().context(err_ctx)?;
+        let major: usize = major.parse().context(err_ctx)?;
+        let minor = parts.next().context(err_ctx)?;
+        let minor: usize = minor.parse().context(err_ctx)?;
+        if (major, minor) > x86_64_default {
+            x86_64_ver = (major, minor);
+        }
+        if (major, minor) > arm64_default {
+            arm64_ver = (major, minor);
+        }
+    }
+    Ok((x86_64_ver, arm64_ver))
+}
+
+#[cfg(test)]
+mod test {
+    use super::macosx_deployment_target;
+
+    #[test]
+    fn test_macosx_deployment_target() {
+        assert_eq!(
+            macosx_deployment_target(None, false).unwrap(),
+            (((10, 7), (11, 0)))
+        );
+        assert_eq!(
+            macosx_deployment_target(None, true).unwrap(),
+            (((10, 9), (11, 0)))
+        );
+        assert_eq!(
+            macosx_deployment_target(Some("10.6"), false).unwrap(),
+            (((10, 7), (11, 0)))
+        );
+        assert_eq!(
+            macosx_deployment_target(Some("10.6"), true).unwrap(),
+            (((10, 9), (11, 0)))
+        );
+        assert_eq!(
+            macosx_deployment_target(Some("10.9"), false).unwrap(),
+            (((10, 9), (11, 0)))
+        );
+        assert_eq!(
+            macosx_deployment_target(Some("11.0.0"), false).unwrap(),
+            (((11, 0), (11, 0)))
+        );
+        assert_eq!(
+            macosx_deployment_target(Some("11.1"), false).unwrap(),
+            (((11, 1), (11, 1)))
+        );
+    }
+}
